==> rad-0.9.0.tar.gz: lib/src/widgets/gesture_detector.dart <==
import 'dart:html';

import 'package:meta/meta.dart';

import 'package:rad/src/core/common/constants.dart';
import 'package:rad/src/core/common/enums.dart';
import 'package:rad/src/core/common/functions.dart';
import 'package:rad/src/core/common/objects/build_context.dart';
import 'package:rad/src/core/common/objects/dom_node_description.dart';
import 'package:rad/src/core/common/objects/key.dart';
import 'package:rad/src/core/common/objects/render_object.dart';
import 'package:rad/src/core/common/types.dart';
import 'package:rad/src/core/services/events/emitted_event.dart';
import 'package:rad/src/core/services/services_registry.dart';
import 'package:rad/src/widgets/abstract/widget.dart';

/// A widget that detects gestures.
///
/// Attempts to recognize gestures that correspond to its non-null callbacks.
///
/// See also:
///
///  * [HitTestBehavior], behaviour of a [GestureDetector]
///
class GestureDetector extends Widget {
  final Widget child;

  final Callback? onTap;
  final Callback? onDoubleTap;
  final EventCallback? onTapEvent;
  final EventCallback? onDoubleTapEvent;

  /// Hit testing behaviour.
  ///
  /// Applies only to tap/double tap events.
  ///
  final HitTestBehavior behaviour;

  final EventCallback? onMouseEnterEvent;
  final EventCallback? onMouseLeaveEvent;

  const GestureDetector({
    Key? key,
    required this.child,
    this.onTap,
    this.onDoubleTap,
    this.onTapEvent,
    this.onDoubleTapEvent,
    this.onMouseEnterEvent,
    this.onMouseLeaveEvent,
    this.behaviour = HitTestBehavior.deferToChild,
  }) : super(key: key);

  @override
  Map<DomEventType, EventCallback?> get widgetEventListeners => {
        DomEventType.mouseEnter: onMouseEnterEvent,
        DomEventType.mouseLeave: onMouseLeaveEvent,
      };

  /*
  |--------------------------------------------------------------------------
  | widget internals
  |--------------------------------------------------------------------------
  */

  @override
  List<Widget> get widgetChildren => [child];

  @nonVirtual
  @override
  String get widgetType => 'GestureDetector';

  // gesture detector creates a dom node(div) because it has to
  // register event listeners on a dom node.

  @nonVirtual
  @override
  DomTagType get correspondingTag => DomTagType.division;

  @override
  createConfiguration() => _GestureDetectorConfiguration(
        behaviour: behaviour,
        onTap: onTap,
        onDoubleTap: onDoubleTap,
        onTapEvent: onTapEvent,
        onDoubleTapEvent: onDoubleTapEvent,
      );

  @nonVirtual
  @override
  isConfigurationChanged(oldConfiguration) => true;

  @nonVirtual
  @override
  createRenderObject(context) => _GestureDetectorRenderObject(
        context: context,
        state: _GestureDetectorState(),
      );
}

/*
|--------------------------------------------------------------------------
| configuration
|--------------------------------------------------------------------------
*/

class _GestureDetectorConfiguration extends WidgetConfiguration {
  final Callback? onTap;
  final Callback? onDoubleTap;
  final EventCallback? onTapEvent;
  final EventCallback? onDoubleTapEvent;

  final HitTestBehavior behaviour;

  const _GestureDetectorConfiguration({
    this.onTap,
    this.onDoubleTap,
    this.onTapEvent,
    this.onDoubleTapEvent,
    this.behaviour = HitTestBehavior.deferToChild,
  });
}

/*
|--------------------------------------------------------------------------
| description(never changes for gesture detector widget)
|--------------------------------------------------------------------------
*/

const _description = DomNodeDescription(
  attributes: {
    Attributes.classAttribute: Constants.classGestureDetector,
  },
);

/*
|--------------------------------------------------------------------------
| render object
|--------------------------------------------------------------------------
*/

class _GestureDetectorRenderObject extends RenderObject {
  final _GestureDetectorState state;

  const _GestureDetectorRenderObject({
    required BuildContext context,
    required this.state,
  }) : super(context);

  @override
  render({
    required covariant _GestureDetectorConfiguration configuration,
  }) {
    var services = ServicesRegistry.instance.getServices(context);
    var widget = services.walker.getWidgetObject(context)!.widget;

    state
      ..frameworkBindDomNode(context.findDomNode())
      ..frameworkBindWidget(widget)
      ..initState();

    return _description;
  }

  @override
  afterWidgetRebind({
    required oldWidget,
    required newWidget,
    required updateType,
  }) {
    state.frameworkReBindWidget(
      oldWidget: oldWidget,
      newWidget: newWidget,
    );
  }

  @override
  beforeUnMount() => state.frameworkDispose();
}

class _GestureDetectorState {
  Element? _domNode;
  Element get domNode => _domNode!;

  GestureDetector? _widget;
  GestureDetector get widget => _widget!;

  final _availableEventListeners = [
    DomEventType.click,
    DomEventType.doubleClick,
  ];

  void initState() {
    _refreshEventListeners(null);
  }

  void didUpdateWidget(GestureDetector oldWidget) {
    _refreshEventListeners(oldWidget);
  }

  void _refreshEventListeners(GestureDetector? oldWidget) {
    for (final eventType in _availableEventListeners) {
      var hasListener = _doWidgetHasListener(eventType, widget);
      var hadListener = false;

      if (null != oldWidget) {
        hadListener = _doWidgetHasListener(eventType, oldWidget);
      }

      //
      // we don't rebind event listeners for tap/doubleTaps.
      // i.e in -> onTap: someVar ? () { A } : () { B }, Callback containing
      // B will never binds as onTap handler
      //

      if (hasListener != hadListener) {
        if (hasListener) {
          _addListener(eventType);
        } else {
          _removeListener(eventType);
        }
      }
    }
  }

  void _removeListeners() {
    for (final eventType in _availableEventListeners) {
      _removeListener(eventType);
    }
  }

  bool _doWidgetHasListener(DomEventType eventType, GestureDetector widget) {
    switch (eventType) {
      case DomEventType.click:
        return null != widget.onTap || null != widget.onTapEvent;

      case DomEventType.doubleClick:
        return null != widget.onDoubleTap || null != widget.onDoubleTapEvent;

      default:
        return false;
    }
  }

  void _addListener(DomEventType eventType) {
    var nativeType = eventType.nativeName;
    var useCapture = HitTestBehavior.opaque == widget.behaviour;

    domNode.addEventListener(nativeType, _handleNative, useCapture);
  }

  void _removeListener(DomEventType eventType) {
    var nativeType = eventType.nativeName;

    domNode.removeEventListener(nativeType, _handleNative);
  }

  void _handleNative(Event event) {
    var emittedEvent = EmittedEvent.fromNativeEvent(event);

    emittedEvent.preventDefault();

    switch (widget.behaviour) {
      case HitTestBehavior.opaque:
      case HitTestBehavior.deferToChild:
        emittedEvent.stopPropagation();

        break;

      case HitTestBehavior.translucent:
        break;
    }

    _dispatch(emittedEvent);
  }

  void _dispatch(EmittedEvent event) {
    Callback? listener;
    EventCallback? eventListener;

    switch (fnMapEventTypeToDomEventType(event.type)) {
      case DomEventType.click:
        listener = widget.onTap;
        eventListener = widget.onTapEvent;

        break;

      case DomEventType.doubleClick:
        listener = widget.onDoubleTap;
        eventListener = widget.onDoubleTapEvent;

        break;

      default:
        break;
    }

    if (null != listener) {
      listener();
    }

    if (null != eventListener) {
      eventListener(event);
    }
  }

  void dispose() {
    _removeListeners();
  }

  /*
  |--------------------------------------------------------------------------
  | for internal use
  |--------------------------------------------------------------------------
  */

  @nonVirtual
  @protected
  void frameworkBindDomNode(Element domNode) {
    _domNode = domNode;
  }

  @nonVirtual
  @protected
  void frameworkBindWidget(Widget widget) {
    _widget = widget as GestureDetector;
  }

  @nonVirtual
  @protected
  void frameworkReBindWidget({
    required Widget oldWidget,
    required Widget newWidget,
  }) {
    _widget = newWidget as GestureDetector;

    didUpdateWidget(oldWidget as GestureDetector);
  }

  @nonVirtual
  @protected
  void frameworkDispose() {
    dispose();
  }
}

==> rad-1.0.0-rc.1.tar.gz: lib/src/widgets/gesture_detector.dart <==
// Copyright (c) 2022, Rad developers. All rights reserved.
// Use of this source code is governed by a BSD-style license that can be
// found in the LICENSE file.

import 'dart:html';

import 'package:meta/meta.dart';

import 'package:rad/src/core/common/abstract/render_element.dart';
import 'package:rad/src/core/common/abstract/watchful_render_element.dart';
import 'package:rad/src/core/common/constants.dart';
import 'package:rad/src/core/common/enums.dart';
import 'package:rad/src/core/common/functions.dart';
import 'package:rad/src/core/common/objects/cache.dart';
import 'package:rad/src/core/common/objects/dom_node_patch.dart';
import 'package:rad/src/core/common/objects/key.dart';
import 'package:rad/src/core/common/types.dart';
import 'package:rad/src/core/services/events/emitted_event.dart';
import 'package:rad/src/widgets/abstract/single_child_widget.dart';
import 'package:rad/src/widgets/abstract/widget.dart';

/// A widget that detects gestures.
///
/// Attempts to recognize gestures that correspond to its non-null callbacks.
///
/// See also:
///
///  * [HitTestBehavior], behaviour of a [GestureDetector]
///
class GestureDetector extends SingleChildWidget {
  final VoidCallback? onTap;
  final VoidCallback? onDoubleTap;
  final EventCallback? onTapEvent;
  final EventCallback? onDoubleTapEvent;

  /// Hit testing behaviour.
  ///
  /// Applies only to tap/double tap events.
  ///
  final HitTestBehavior behaviour;

  final EventCallback? onMouseEnterEvent;
  final EventCallback? onMouseLeaveEvent;

  const GestureDetector({
    Key? key,
    this.onTap,
    this.onDoubleTap,
    this.onTapEvent,
    this.onDoubleTapEvent,
    this.onMouseEnterEvent,
    this.onMouseLeaveEvent,
    this.behaviour = HitTestBehavior.deferToChild,
    required Widget child,
  }) : super(key: key, child: child);

  @override
  Map<DomEventType, EventCallback?> get widgetEventListeners {
    if (null == onMouseEnterEvent && null == onMouseLeaveEvent) {
      return ccImmutableEmptyMapOfEventListeners;
    }

    return {
      DomEventType.mouseEnter: onMouseEnterEvent,
      DomEventType.mouseLeave: onMouseLeaveEvent,
    };
  }

  /*
  |--------------------------------------------------------------------------
  | widget internals
  |--------------------------------------------------------------------------
  */

  @nonVirtual
  @override
  String get widgetType => 'GestureDetector';

  // gesture detector creates a dom node(div) because it has to
  // register event listeners on a dom node.

  @nonVirtual
  @override
  DomTagType get correspondingTag => DomTagType.division;

  @override
  shouldUpdateWidget(Widget oldWidget) => true;

  @nonVirtual
  @override
  createRenderElement(parent) => GestureDetectorRenderElement(this, parent);
}

/*
|--------------------------------------------------------------------------
| description(never changes for gesture detector widget)
|--------------------------------------------------------------------------
*/

const _description = DomNodePatch(
  attributes: {
    Attributes.classAttribute: Constants.classGestureDetector,
  },
);

/*
|--------------------------------------------------------------------------
| render element
|--------------------------------------------------------------------------
*/

/// Gesture detector's render element.
///
class GestureDetectorRenderElement extends WatchfulRenderElement {
  /// Associated state.
  ///
  final _GestureDetectorState state;

  GestureDetectorRenderElement(
    GestureDetector widget,
    RenderElement parent,
  )   : state = _GestureDetectorState(),
        _widgetChildren = [widget.child],
        super(widget, parent);

  @override
  List<Widget> get widgetChildren => _widgetChildren;
  List<Widget> _widgetChildren;

  @mustCallSuper
  @override
  init() {
    state
      ..frameworkBindDomNode(domNode!)
      ..frameworkBindWidget(widget)
      ..initState();
  }

  @override
  render({required widget}) => _description;

  @override
  void afterWidgetRebind({
    required oldWidget,
    required covariant SingleChildWidget newWidget,
    required updateType,
  }) {
    _widgetChildren = [newWidget.child];

    state.frameworkRebindWidget(
      oldWidget: oldWidget,
      newWidget: newWidget,
    );
  }

  @override
  afterUnMount() => state.frameworkDispose();
}

/// Gesture detector's state.
///
class _GestureDetectorState {
  Element? _domNode;
  Element get domNode => _domNode!;

  GestureDetector? _widget;
  GestureDetector get widget => _widget!;

  final _availableEventListeners = [
    DomEventType.click,
    DomEventType.doubleClick,
  ];

  void initState() {
    _refreshEventListeners(null);
  }

  void didUpdateWidget(GestureDetector oldWidget) {
    _refreshEventListeners(oldWidget);
  }

  void _refreshEventListeners(GestureDetector? oldWidget) {
    for (final eventType in _availableEventListeners) {
      var hasListener = _doWidgetHasListener(eventType, widget);
      var hadListener = false;

      if (null != oldWidget) {
        hadListener = _doWidgetHasListener(eventType, oldWidget);
      }

      //
      // we don't rebind event listeners for tap/doubleTaps.
      // i.e in -> onTap: someVar ? () { A } : () { B }, Callback containing
      // B will never binds as onTap handler
      //

      if (hasListener != hadListener) {
        if (hasListener) {
          _addListener(eventType);
        } else {
          _removeListener(eventType);
        }
      }
    }
  }

  void _removeListeners() {
    for (final eventType in _availableEventListeners) {
      _removeListener(eventType);
    }
  }

  bool _doWidgetHasListener(DomEventType eventType, GestureDetector widget) {
    switch (eventType) {
      case DomEventType.click:
        return null != widget.onTap || null != widget.onTapEvent;

      case DomEventType.doubleClick:
        return null != widget.onDoubleTap || null != widget.onDoubleTapEvent;

      default:
        return false;
    }
  }

  void _addListener(DomEventType eventType) {
    var nativeType = eventType.nativeName;
    var useCapture = HitTestBehavior.opaque == widget.behaviour;

    domNode.addEventListener(nativeType, _handleNative, useCapture);
  }

  void _removeListener(DomEventType eventType) {
    var nativeType = eventType.nativeName;

    domNode.removeEventListener(nativeType, _handleNative);
  }

  void _handleNative(Event event) {
    var emittedEvent = EmittedEvent.fromNativeEvent(event);

    emittedEvent.preventDefault();

    switch (widget.behaviour) {
      case HitTestBehavior.opaque:
      case HitTestBehavior.deferToChild:
        emittedEvent.stopPropagation();

        break;

      case HitTestBehavior.translucent:
        break;
    }

    _dispatch(emittedEvent);
  }

  void _dispatch(EmittedEvent event) {
    VoidCallback? listener;
    EventCallback? eventListener;

    switch (fnMapEventTypeToDomEventType(event.type)) {
      case DomEventType.click:
        listener = widget.onTap;
        eventListener = widget.onTapEvent;

        break;

      case DomEventType.doubleClick:
        listener = widget.onDoubleTap;
        eventListener = widget.onDoubleTapEvent;

        break;

      default:
        break;
    }

    if (null != listener) {
      listener();
    }

    if (null != eventListener) {
      eventListener(event);
    }
  }

  void dispose() {
    _removeListeners();
  }

  /*
  |--------------------------------------------------------------------------
  | for internal use
  |--------------------------------------------------------------------------
  */

  @nonVirtual
  @protected
  void frameworkBindDomNode(Element domNode) {
    assert(null == _domNode, 'DomNode is already bound');

    _domNode = domNode;
  }

  @nonVirtual
  @protected
  void frameworkBindWidget(Widget widget) {
    assert(null == _widget, 'Widget is already bound');

    _widget = widget as GestureDetector;
  }

  @nonVirtual
  @protected
  void frameworkRebindWidget({
    required Widget oldWidget,
    required Widget newWidget,
  }) {
    assert(null != _widget, 'Widget is not bound yet');

    _widget = newWidget as GestureDetector;
    didUpdateWidget(oldWidget as GestureDetector);
  }

  @nonVirtual
  @protected
  void frameworkDispose() {
    dispose();
  }
}
